// ----------------------------------------------------------------------------
// (C) 2002 
//     carrara engineering
//     Switzerland
//     http://www.carrara.ch
//     mailto:info@carrara.ch
// ----------------------------------------------------------------------------
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author make no representations about the suitability 
//     of this software for any purpose. It is provided "as is" 
//     without express or implied warranty.
// ----------------------------------------------------------------------------


#include <iostream.h>

// ----------------------------------------------------------------------------

namespace PrePostAction1
{

void preAction() {cout << "preAction" << endl;}
void postAction() {cout << "postAction" << endl;}

class User
{
  public:
    User(const char* const i_pcString)
    : m_pcString (i_pcString)
    {}

    void fn1(const char* const i_pcString)
    {
      cout << "1) " << m_pcString << ": " << i_pcString << endl;
    }
    void fn2(const char* const i_pcString)
    {
      cout << "2) " << m_pcString << ": " << i_pcString << endl;
    }
  private:
    const char* const m_pcString;
};

void test()
{
  User aUser("User Object");

  preAction();
  aUser.fn1("function call");
  postAction();

  preAction();
  aUser.fn2("function call");
  postAction();
}

}

// ----------------------------------------------------------------------------

namespace PrePostAction2
{

void preAction() {cout << "preAction" << endl;}
void postAction() {cout << "postAction" << endl;}

class User
{
    typedef void (*T_pFn)();
  public:
    User(const char* const i_pcString, T_pFn i_pFnPre, T_pFn i_pFnPost)
    : m_pcString (i_pcString)
    , m_pFnPre   (i_pFnPre)
    , m_pFnPost  (i_pFnPost)
    {}

    void fn1(const char* const i_pcString)
    {
      (*m_pFnPre)();
      cout << "1) " << m_pcString << ": " << i_pcString << endl;
      (*m_pFnPost)();
    }

    void fn2(const char* const i_pcString)
    {
      (*m_pFnPre)();
      cout << "2) " << m_pcString << ": " << i_pcString << endl;
      (*m_pFnPost)();
    }

  private:
    const char* const m_pcString;
    T_pFn m_pFnPre;
    T_pFn m_pFnPost;
};

void test()
{
  User aUser("User Object",&preAction,&postAction);

  aUser.fn1("function call");
  aUser.fn2("function call");

}

}

// ----------------------------------------------------------------------------

namespace PrePostAction3
{

class User
{
  public:
    User(const char* const i_pcString)
    : m_pcString (i_pcString)
    {}

    void fn1(const char* const i_pcString)
    {
      cout << "1) " << m_pcString << ": " << i_pcString << endl;
    }
    void fn2(const char* const i_pcString)
    {
      cout << "2) " << m_pcString << ": " << i_pcString << endl;
    }
  private:
    const char* const m_pcString;
};

class PreAction
{
  public:
    PreAction(User& i_raUser)
    : m_raUser (i_raUser)
    {}

    User* operator->()
    {
      preAction();
      return &m_raUser;
    }

  private: 
    User& m_raUser;

    void preAction() {cout << "preAction" << endl;}

};

void test()
{
  User aUser("User Object");

  PreAction(aUser)->fn1("function call");
  PreAction(aUser)->fn2("function call");
}

}

// ----------------------------------------------------------------------------

namespace PrePostAction4
{

class User
{
  public:
    User(const char* const i_pcString)
    : m_pcString (i_pcString)
    {}

    void fn1(const char* const i_pcString)
    {
      cout << "1) " << m_pcString << ": " << i_pcString << endl;
    }
    void fn2(const char* const i_pcString)
    {
      cout << "2) " << m_pcString << ": " << i_pcString << endl;
    }
  private:
    const char* const m_pcString;
};


class PrePostAction
{
  private:
    class Caller
    {
      public:
        Caller(User& i_raUser)
        : m_raUser (i_raUser)
        {
          preAction();
        }

        ~Caller()
        {
          postAction();
        }

        User* operator->()
        {
          return &m_raUser;
        }

      private: 
        User& m_raUser;

        void preAction() {cout << "preAction" << endl;}
        void postAction() {cout << "postAction" << endl;}
    };
  public:
    PrePostAction(User& i_raUser)
    : m_raUser (i_raUser)
    {}

    Caller operator->()
    {
      return Caller(m_raUser);
    }

  private: 
    User& m_raUser;

};

void test()
{
  User aUser("User Object");

  PrePostAction(aUser)->fn1("function call");
  PrePostAction(aUser)->fn2("function call");
}

}

// ----------------------------------------------------------------------------

namespace PrePostAction5
{

class User
{
  public:
    User(const char* const i_pcString)
    : m_pcString (i_pcString)
    {}

    void fn1(const char* const i_pcString)
    {
      cout << "1) " << m_pcString << ": " << i_pcString << endl;
    }
    void fn2(const char* const i_pcString)
    {
      cout << "2) " << m_pcString << ": " << i_pcString << endl;
    }
  private:
    const char* const m_pcString;
};

template <class T>
class PrePostAction
{
  private:
    class Caller
    {
      public:
        Caller(T& i_raObj)
        : m_raObj (i_raObj)
        {
          preAction();
        }

        ~Caller()
        {
          postAction();
        }

        T* operator->()
        {
          return &m_raObj;
        }

      private: 
        T& m_raObj;

        void preAction() {cout << "preAction" << endl;}
        void postAction() {cout << "postAction" << endl;}
    };

  public:
    PrePostAction(T& i_raObj)
    : m_raObj (i_raObj)
    {}

    Caller operator->()
    {
      return Caller(m_raObj);
    }

  private: 
    T& m_raObj;

};

void test()
{
  User aUser("User Object");

  PrePostAction<User>(aUser)->fn1("function call");
  PrePostAction<User>(aUser)->fn2("function call");
}

}

// ----------------------------------------------------------------------------

namespace PrePostAction6
{

class User
{
  public:
    User(const char* const i_pcString)
    : m_pcString (i_pcString)
    {}

    void fn1(const char* const i_pcString)
    {
      cout << "1) " << m_pcString << ": " << i_pcString << endl;
    }
    void fn2(const char* const i_pcString)
    {
      cout << "2) " << m_pcString << ": " << i_pcString << endl;
    }
  private:
    const char* const m_pcString;
};

template <class T, class S>
class PrePostAction
{
  private:
    class Caller
    {
      public:
        Caller(T& i_raObj)
        : m_raObj (i_raObj)
        {
          m_aStrategy.preAction();
        }

        ~Caller()
        {
          m_aStrategy.postAction();
        }

        T* operator->()
        {
          return &m_raObj;
        }

      private: 
        T& m_raObj;
        S m_aStrategy;

    };

  public:
    PrePostAction(T& i_raObj)
    : m_raObj (i_raObj)
    {}

    Caller operator->()
    {
      return Caller(m_raObj);
    }

  private: 
    T& m_raObj;

};

class PrePostStrategy
{
  public:
    void preAction() {cout << "preAction" << endl;}
    void postAction() {cout << "postAction" << endl;}
};

void test()
{
  User aUser("User Object");

  PrePostAction<User,PrePostStrategy>(aUser)->fn1("function call");
  PrePostAction<User,PrePostStrategy>(aUser)->fn2("function call");
}

}

// ----------------------------------------------------------------------------

namespace PrePostAction7
{

class User
{
  public:
    User(const char* const i_pcString)
    : m_pcString (i_pcString)
    {}

    void fn1(const char* const i_pcString)
    {
      cout << "1) " << m_pcString << ": " << i_pcString << endl;
    }
    void fn2(const char* const i_pcString)
    {
      cout << "2) " << m_pcString << ": " << i_pcString << endl;
    }
  private:
    const char* const m_pcString;
};

template <class T, class S>
class PrePostAction
{
  private:
    class Caller
    {
      public:
        Caller(T& i_raObj)
        : m_raObj (i_raObj)
        {
          m_aStrategy.preAction();
        }

        ~Caller()
        {
          m_aStrategy.postAction();
        }

        T* operator->()
        {
          return &m_raObj;
        }

      private: 
        T& m_raObj;
        S m_aStrategy;

    };

  public:
    PrePostAction(T& i_raObj)
    : m_raObj (i_raObj)
    {}

    Caller operator->()
    {
      return Caller(m_raObj);
    }

  private: 
    T& m_raObj;

};

class PrePostStrategy
{
  public:
    void preAction() {cout << "preAction" << endl;}
    void postAction() {cout << "postAction" << endl;}
};

template <class S, class T>
PrePostAction<T,S> Action(T& i_raObj)
{
  return PrePostAction<T,S>(i_raObj);
}

void test()
{
  User aUser("User Object");

  Action<PrePostStrategy>(aUser)->fn1("function call");
  Action<PrePostStrategy>(aUser)->fn2("function call");
}


}


// ----------------------------------------------------------------------------

namespace PrePostAction8
{

class User
{
  public:
    User(const char* const i_pcString)
    : m_pcString (i_pcString)
    {}

    void fn1(const char* const i_pcString)
    {
      cout << "1) " << m_pcString << ": " << i_pcString << endl;
    }
    void fn2(const char* const i_pcString)
    {
      cout << "2) " << m_pcString << ": " << i_pcString << endl;
    }
  private:
    const char* const m_pcString;
};

template <class T, class S>
class PrePostAction
{
  private:
    class Caller
    {
      public:
        Caller(T& i_raObj, S& i_raStrategy)
        : m_raObj      (i_raObj)
        , m_raStrategy (i_raStrategy)
        {
          m_raStrategy.preAction();
        }

        ~Caller()
        {
          m_raStrategy.postAction();
        }

        T* operator->()
        {
          return &m_raObj;
        }

      private: 
        T& m_raObj;
        S& m_raStrategy;

    };

  public:
    PrePostAction(T& i_raObj);

    PrePostAction(T& i_raObj, const S& i_raStrategy);

    Caller operator->()
    {
      return Caller(m_raObj,m_aStrategy);
    }

  private: 
    T& m_raObj;
    S m_aStrategy;

};

template <class T, class S>
PrePostAction<T,S>::PrePostAction(T& i_raObj)
: m_raObj (i_raObj)
{}

template <class T, class S>
PrePostAction<T,S>::PrePostAction(T& i_raObj, const S& i_raStrategy)
: m_raObj     (i_raObj)
, m_aStrategy (i_raStrategy)
{}


class PrePostStrategy
{
  public:
    void preAction() {cout << "preAction" << endl;}
    void postAction() {cout << "postAction" << endl;}
};


class PrePostCountStrategy
{
  public:
    
    PrePostCountStrategy(int& i_raCallCounter)
    : m_raCallCounter (i_raCallCounter)
    {}

    PrePostCountStrategy(const PrePostCountStrategy& i_raStrategy)
    : m_raCallCounter (i_raStrategy.m_raCallCounter)
    {}

    void preAction() 
    {
      ++m_raCallCounter;
      cout << "preAction" << endl;
    }
    void postAction() 
    {
      cout << "postAction" << endl;
    }

  private:

    int& m_raCallCounter;
};



template <class S, class T>
PrePostAction<T,S> Action(T& i_raObj)
{
  return PrePostAction<T,S>(i_raObj);
}

template <class S, class T>
PrePostAction<T,S> Action(T& i_raObj, const S& i_raStrategy)
{
  return PrePostAction<T,S>(i_raObj, i_raStrategy);
}

void test()
{
  int nCounter = 0;
  User aUser("User Object");

  Action<PrePostStrategy>(aUser)->fn1("function call");
  Action<PrePostStrategy>(aUser)->fn2("function call");

  Action(aUser, PrePostCountStrategy(nCounter))->fn1("function call");
  Action(aUser, PrePostCountStrategy(nCounter))->fn2("function call");

  Action<PrePostCountStrategy>(aUser, nCounter)->fn1("function call");
  Action<PrePostCountStrategy>(aUser, nCounter)->fn2("function call");

  cout << "CallCounter: " << nCounter << endl;

}


}

// ----------------------------------------------------------------------------

namespace PrePostAction9
{

class User
{
  public:
    User(const char* const i_pcString)
    : m_pcString (i_pcString)
    {}

    void fn1(const char* const i_pcString)
    {
      cout << "1) " << m_pcString << ": " << i_pcString << endl;
    }
    void fn2(const char* const i_pcString)
    {
      cout << "2) " << m_pcString << ": " << i_pcString << endl;
    }
  private:
    const char* const m_pcString;
};

template <class T, class S>
class PrePostAction
{
  private:
    class Caller
    {
      public:
        Caller(T& i_raObj, S& i_raStrategy)
        : m_raObj      (i_raObj)
        , m_raStrategy (i_raStrategy)
        {
          m_raStrategy.preAction();
        }

        ~Caller()
        {
          m_raStrategy.postAction();
        }

        T* operator->()
        {
          return &m_raObj;
        }

      private: 
        T& m_raObj;
        S& m_raStrategy;

    };

  public:
    PrePostAction(T& i_raObj);

    PrePostAction(T& i_raObj, const S& i_raStrategy);

    Caller operator->()
    {
      return Caller(m_raObj,m_aStrategy);
    }

  private: 
    T& m_raObj;
    S m_aStrategy;

};

template <class T, class S>
PrePostAction<T,S>::PrePostAction(T& i_raObj)
: m_raObj (i_raObj)
{}

template <class T, class S>
PrePostAction<T,S>::PrePostAction(T& i_raObj, const S& i_raStrategy)
: m_raObj     (i_raObj)
, m_aStrategy (i_raStrategy)
{}


class Semaphore
{
  public:
    void aquire() {cout << "Semaphore::aquire()" << endl;}
    void release() {cout << "Semaphore::release()" << endl;}
};


class SemaphoreStrategy
{
  public:
    
    SemaphoreStrategy(Semaphore& i_raSemaphore)
    : m_raSemaphore (i_raSemaphore)
    {}

    SemaphoreStrategy(const SemaphoreStrategy& i_raStrategy)
    : m_raSemaphore (i_raStrategy.m_raSemaphore)
    {}

    void preAction() 
    {
      m_raSemaphore.aquire();
    }
    void postAction() 
    {
      m_raSemaphore.release();
    }

  private:

    Semaphore& m_raSemaphore;
};



template <class S, class T>
PrePostAction<T,S> Action(T& i_raObj)
{
  return PrePostAction<T,S>(i_raObj);
}

template <class S, class T>
PrePostAction<T,S> Action(T& i_raObj, const S& i_raStrategy)
{
  return PrePostAction<T,S>(i_raObj, i_raStrategy);
}

void test()
{
  Semaphore aSemaphore;
  
  User aUser("User Object");

  Action<SemaphoreStrategy>(aUser, aSemaphore)->fn1("function call");
  Action<SemaphoreStrategy>(aUser, aSemaphore)->fn2("function call");

}


}

// ----------------------------------------------------------------------------

main()
{
  cout << "--------------------------- Test 1 ---------------------------" << endl;
  PrePostAction1::test();

  cout << "--------------------------- Test 2 ---------------------------" << endl;
  PrePostAction2::test();

  cout << "--------------------------- Test 3 ---------------------------" << endl;
  PrePostAction3::test();

  cout << "--------------------------- Test 4 ---------------------------" << endl;
  PrePostAction4::test();

  cout << "--------------------------- Test 5 ---------------------------" << endl;
  PrePostAction5::test();

  cout << "--------------------------- Test 6 ---------------------------" << endl;
  PrePostAction6::test();

  cout << "--------------------------- Test 7 ---------------------------" << endl;
  PrePostAction7::test();

  cout << "--------------------------- Test 8 ---------------------------" << endl;
  PrePostAction8::test();

  cout << "--------------------------- Test 9 ---------------------------" << endl;
  PrePostAction9::test();

  return 0;
}
